from typing import Type, Sequence, Any, Optional

import os
import subprocess
import tempfile
import re
import asyncio
import threading
import time
import shutil

import logging
logger = logging.getLogger("[clickhouse_test]")

from importlib.resources import files

CLICKHOUSE_SERVER = os.environ.get('CLICKHOUSE_SERVER', "/usr/bin/clickhouse-server")
CLICKHOUSE_SERVER_CONFIG = os.environ.get('CLICKHOUSE_SERVER_CONFIG')
CLICKHOUSE_SERVER_USERS = os.environ.get('CLICKHOUSE_SERVER_USERS')
CLICKHOUSE_SERVER_CLEANUP = os.environ.get('CLICKHOUSE_SERVER_CLEANUP')

CLICKHOUSE_LISTENS = re.compile(r"^.*Application: Ready for connections\.$")
START_TIMEOUT = 5.0
LINE_FEED_TIMEOUT = 1.0
COROUTINE_SLEEP = 0.001
THREAD_SLEEP = COROUTINE_SLEEP

def log_stdout(p: subprocess.Popen):
    assert p.stdout
    outline = p.stdout.readline()
    while p.poll() is None or outline:
        if outline:
            logger.debug(outline)
        else:
            time.sleep(THREAD_SLEEP)
        outline = p.stdout.readline()

class ClickHouseTestServer:
    async def check_listener(self) -> bool:
        assert self.process.stdout
        os.set_blocking(self.process.stdout.fileno(), False)

        while self.process.poll() is None:
            line = self.process.stdout.readline()
            if line is None or line == '':
                await asyncio.sleep(COROUTINE_SLEEP)
            else:
                logger.debug(line)
                if CLICKHOUSE_LISTENS.fullmatch(line.strip()):
                    return True

        logger.error("Unexpected ClickHouse server termination: retcode=%i", self.process.returncode)

        return False

    async def wait_checker(self) -> bool:
        return await asyncio.wait_for(self.check_listener(),
                                      timeout=START_TIMEOUT)

    def wait_listener(self) -> bool:
        return asyncio.run(self.wait_checker())

    def run_clickhouse(self) -> None:
        self.workdir = tempfile.mkdtemp(prefix="clickhouse_test.")

        logger.info("Setup configuration files...")

        if not CLICKHOUSE_SERVER_CONFIG:
            server_config = os.path.join(self.workdir, 'clickhouse-server-config.xml')
            with open(server_config, "w") as conf_file:
                conf_file.write(files('clickhouse_test').joinpath('clickhouse-server-config.xml').read_text())
        else:
            server_config = CLICKHOUSE_SERVER_CONFIG

        if not CLICKHOUSE_SERVER_USERS:
            users_config = os.path.join(self.workdir, 'clickhouse-server-users.xml')
            with open(users_config, "w") as conf_file:
                conf_file.write(files('clickhouse_test').joinpath('clickhouse-server-users.xml').read_text())
        else:
            users_config = CLICKHOUSE_SERVER_USERS

        self.env = {
            'LOGFILE': os.path.join(self.workdir, 'clickhouse-server.log'),
            'ERROR_LOGFILE': os.path.join(self.workdir, 'clickhouse-server-err.log'),
            'WORKDIR': os.path.join(self.workdir, ''),
            'TEMPDIR': os.path.join(self.workdir, 'temp', ''),
            'USERDIR': os.path.join(self.workdir, 'user', ''),
            'USERS_XML': users_config,
            'ACCESSDIR': os.path.join(self.workdir, 'access', '')
        }

        logger.info("Starting ClickHouse server...")

        self.process = subprocess.Popen([CLICKHOUSE_SERVER,
                                        f"--config={server_config}"],
                                        env=self.env, cwd=self.workdir,
                                        stderr=subprocess.STDOUT, stdout=subprocess.PIPE,
                                        universal_newlines=True)

        logger.info("ClickHouse server started (PID=%i)." % self.process.pid)

        if self.wait_listener():
            logger.info("ClickHouse server is listening for connections...")
        else:
            raise AssertionError("Something is going wrong: the ClickHouse server isn't listening for connections!")

        threading.Thread(target=log_stdout, args=(self.process,), daemon=True).start()

    def stop_clickhouse(self) -> None:
        if self.process:
            logger.info("Stopping ClickHouse server...")
            self.process.terminate()
            self.process.wait()
            logger.info("ClickHouse server stopped.")

        if CLICKHOUSE_SERVER_CLEANUP and CLICKHOUSE_SERVER_CLEANUP != "0":
            shutil.rmtree(self.workdir)
            logger.info("Working directory removed.")
            logger.info("To run this ClickHouse server again first repeat the test with CLICKHOUSE_SERVER_CLEANUP unset or set to 0.")
        else:
            logger.info("The workdir was %s" % self.workdir)
            logger.info("To run this ClickHouse server again use: %s" % \
                        " ".join("%s=%s" % (var, self.env[var]) for var in self.env.keys()) + \
                        " " + \
                        " ".join(str(a) for a in self.process.args) if isinstance(self.process.args, Sequence) else str(self.process.args))

    def __enter__(self) -> 'ClickHouseTestServer':
        try:
            self.run_clickhouse()
            return self
        except:
            self.stop_clickhouse()
            raise

    def __exit__(self, exc_type: Type[Exception],
                 exc_value: Exception, traceback) -> None:
        self.stop_clickhouse()
        assert self.process.returncode == 0
